@@ -29,6 +29,12 @@ playDevOff <- function(playState = playDevCur())
     cleanupStateEnv()
 }
 
+playGetIDs <- function(playState = playDevCur(), labels = FALSE)
+{
+    ids <- do.call(rbind, playState$ids)$which
+    if (labels) playState$labels[ids] else ids
+}
+
 print.playState <- function(x, ...)
 {
     stopifnot(inherits(x, "playState"))
@@ -77,15 +83,15 @@ callArg <- function(playState, arg, expr, data = NULL)
     if (is.symbol(arg)) arg <- as.character(arg)
     if (is.null(arg)) return()
     ## instantiate implicit lists as language objects (so deparse is pretty)
-    if (is.language(arg)) {
-        xbits <- strsplit(deparseOneLine(arg), "$", fixed=TRUE)[[1]]
-        for (i in seq_len(length(xbits) - 1)) {
-            implicitList <- parse(text=paste(xbits[1:i], collapse="$"))[[1]]
-            if (is.null(callArg(playState, implicitList))) {
-                callArg(playState, implicitList) <- quote(list())
-            }
-        }
-    }
+#    if (is.language(arg)) {
+#        xbits <- strsplit(deparseOneLine(arg), "$", fixed=TRUE)[[1]]
+#        for (i in seq_len(length(xbits) - 1)) {
+#            implicitList <- parse(text=paste(xbits[1:i], collapse="$"))[[1]]
+#            if (is.null(callArg(playState, implicitList))) {
+#                callArg(playState, implicitList) <- quote(list())
+#            }
+#        }
+#    }
     getx <- if (is.numeric(arg)) paste('[[', arg+1, ']]', sep="")
     else if (is.character(arg)) paste("$", arg, sep="")
     else paste("$", deparseOneLine(arg), sep="")
@@ -209,33 +215,78 @@ rawXYLim <- function(playState, space="plot")
     playState
 }
 
+is.somesortoftime <- function(x) {
+  inherits(x, "Date") ||
+  inherits(x, "POSIXt") ||
+  inherits(x, "yearmon") ||
+  inherits(x, "yearqtr")
+}
+
 setRawXYLim <- function(playState, x, x.or.y=c("x", "y"))
 {
     playDevSet(playState)
     x.or.y <- match.arg(x.or.y)
-    ## convert back from log scale if required
-    x <- spaceCoordsToDataCoordsXY(playState, x, x.or.y=x.or.y)
     ## round digits conservatively
-    x <- round(x, digits=6)
+    x <- signif(x, digits=8)
     if (playState$is.lattice) {
         ## TODO: this really sucks
         x.panel <- xyData(playState, space="page")[[x.or.y]]
-        ## convert new scale to appropriate date time class if required
-        ##if (inherits(x.panel, "Date") || inherits(x.panel, "POSIXt")) {
-        ##	mostattributes(x) <- attributes(x.panel)
-        ##}
-        ## convert new scale to factor levels if required
+        ## set factor labels explicitly, othewise is coerced to numeric
         if (is.factor(x.panel)) {
             if (is.null(callArg(playState, expr=scales[[x.or.y]]$labels))) {
                 callArg(playState, expr=scales[[x.or.y]]$labels) <- levels(x.panel)
                 callArg(playState, expr=scales[[x.or.y]]$at) <- 1:nlevels(x.panel)
             }
         }
+        else if (is.somesortoftime(x.panel)) {
+          class(x) <- class(x.panel)
+          if (inherits(x.panel, "Date"))
+            x <- call("as.Date", format(x))
+          if (inherits(x.panel, "POSIXct"))
+            x <- call("as.POSIXct", format(x))
+          if (inherits(x.panel, "yearmon"))
+            x <- call("as.yearmon", format(as.Date(x)))
+          if (inherits(x.panel, "yearqtr"))
+            x <- call("as.yearqtr", format(as.Date(x)))
+        }
+        else {
+          ## numeric
+          isExtended <- switch(x.or.y,
+                               x = playState$trellis$x.scales$axs == "r",
+                               y = playState$trellis$y.scales$axs == "r")
+          f <- lattice.getOption("axis.padding")$numeric
+          if (isExtended) x <- shrinkrange(x, f=f)
+          ## round digits conservatively
+          x <- signif(x, digits=8)
+        }
     }
+    else if (!is.null(playState$viewport)) {
+      ## non-lattice grid plot
+      ## (do not know if the range is extended or not).
+    }
+    else {
+      ## base graphics plot
+      isExtended <- switch(x.or.y,
+                           x = (par("xaxs") == "r"),
+                           y = (par("yaxs") == "r"))
+      if (isExtended) x <- shrinkrange(x, f=0.04)
+      ## round digits conservatively
+      x <- signif(x, digits=8)
+    }
+    ## convert back from log scale if required
+    x <- spaceCoordsToDataCoordsXY(playState, x, x.or.y=x.or.y)
     if (x.or.y == "x") callArg(playState, "xlim") <- x
     if (x.or.y == "y") callArg(playState, "ylim") <- x
 }
 
+shrinkrange <- function(r, f = 0.1)
+{
+  stopifnot(length(r) == 2)
+  orig.d <- diff(r) / (1 + 2*f)
+  orig.r <- r - c(-f, f) * orig.d
+  orig.r
+}
+
 ## note space="page" means the root viewport
 playDo <- function(playState, expr, space="plot", clip.off=FALSE)
 {
@@ -247,7 +298,7 @@ playDo <- function(playState, expr, space="plot", clip.off=FALSE)
         ## user / plot coordinates
         if (!is.null(playState$viewport)) {
             ## grid graphics plot
-            depth <- downViewport(playState$viewport[[space]])
+            depth <- try(downViewport(playState$viewport[[space]]))
             if (inherits(depth, "try-error")) {
                 stop(paste("Viewport", playState$viewport, "not found"))
             }
@@ -271,7 +322,7 @@ playDo <- function(playState, expr, space="plot", clip.off=FALSE)
             myCol <- col(packets)[whichOne]
             myRow <- row(packets)[whichOne]
             myVp <- trellis.vpname("panel", myCol, myRow, clip.off=clip.off)
-            depth <- downViewport(myVp)
+            depth <- try(downViewport(myVp))
             on.exit(upViewport(depth), add=TRUE)
             ## TODO: should focus panel or just go to viewport?
             ## (if focus, will destroy any previous focus)
@@ -286,6 +337,7 @@ playDo <- function(playState, expr, space="plot", clip.off=FALSE)
             on.exit(upViewport(depth), add=TRUE)
         }
     }
+    if (inherits(depth, "try-error")) return()
     if (!is.null(cur.vp)) on.exit(downViewport(cur.vp), add=TRUE)
     ## do the stuff and return the result
     ##if (is.list(stuff)) lapply(stuff, eval, parent.frame(), playState$env)
@@ -424,6 +476,7 @@ playClickOrDrag <-
              shape=c("rect", "line"),
              scales=c("x", "y"))
 {
+    playDevSet(playState)
     foo <- handleClickOrDrag(playState$widgets$drawingArea,
                              x0=x0, y0=y0, shape=shape, scales=scales)
     if (is.null(foo)) return(NULL)
@@ -487,22 +540,25 @@ handleClickOrDrag <-
         wd <- xx[2] - xx[1]
         ht <- yy[2] - yy[1]
         switch(env$shape,
-               line=gdkDrawLine(event$window, gc=env$gc,
-               x1=env$px0$x, y1=env$px0$y,
-               x2=env$px00$x, y2=env$px00$y),
-               rect=gdkDrawRectangle(event$window, gc=env$gc, filled=FALSE,
-               x=xx[1], yy[1], width=wd, height=ht)
+               line = gdkDrawLine(event$window, gc=env$gc,
+                 x1=env$px0$x, y1=env$px0$y,
+                 x2=env$px00$x, y2=env$px00$y),
+               rect = gdkDrawRectangle(event$window, gc=env$gc, filled=FALSE,
+                 x=xx[1], yy[1], width=wd, height=ht)
                )
         return(TRUE) ## stop event here
     }
     tmpSigE <- gSignalConnect(da, "expose-event", expose_handler, data=environment())
     tmpSigR <- gSignalConnect(da, "button-release-event", release_handler, data=environment())
+    init_time <- proc.time()[3]
     repeat {
         ## px1 is the final drag location, set by event handler
         if (exists("px1", inherits=FALSE)) break
         px00.prev <- px00
         px00 <- da$window$getPointer()
-        if (is.null(px00$retval)) break
+        ## check that pointer is inside the window
+        ## TODO -- fails on linux?
+#        if (is.null(px00$retval)) break
         if ((as.flag(px00$mask) & GdkModifierType["button1-mask"]) == 0) {
             ## mouse button was released
             px1 <- px00
@@ -517,6 +573,10 @@ handleClickOrDrag <-
         ht <- yy[2] - yy[1] + 2
         da$window$invalidateRect(list(x=xx[1], y=yy[1], width=wd, height=ht),
                                  invalidate.children=FALSE)
+        ## try to force redraw
+        gdkWindowProcessAllUpdates()
+        while (gtkEventsPending()) gtkMainIterationDo(blocking=FALSE)
+
     }
     gSignalHandlerDisconnect(da, tmpSigR)
     gSignalHandlerDisconnect(da, tmpSigE)
@@ -529,9 +589,11 @@ handleClickOrDrag <-
     da$window$invalidateRect(invalidate.children=FALSE)
     if (!exists("px1", inherits=FALSE)) return(NULL)
     dc <- list(x=c(px0$x, px1$x), y=c(px0$y, px1$y))
-    ## was it a click or a drag? (threshold 3 pixels)
-    ## TODO: better to use timing
-    is.click <- ((abs(diff(dc$x)) <= 3) && (abs(diff(dc$y)) <= 3))
+    ## was it a click or drag? (click = no slower than 1/4 second)
+    is.click <- (proc.time()[3] - init_time) <= 0.25
+    ## alternative criteria for click: moved less than 10 pixels
+    is.click <- is.click ||
+                ((abs(diff(dc$x)) < 10) && (abs(diff(dc$y)) < 10))
     ndc <- list(x=dc$x / da.w, y=(da.h - dc$y) / da.h)
     list(dc=dc, ndc=ndc, is.click=is.click, modifiers=modifiers)
 }
@@ -578,13 +640,8 @@ xyData <- function(playState, space="plot")
         }
         return(foo)
     }
-    if (playState$is.ggplot) {
-        datarg <- callArg(playState, "data")
-        x <- callArg(playState, 1, data=datarg)
-        y <- callArg(playState, "y", data=datarg)
-        xy <- xy.coords_with_class(playState, x, y)
-        return(xy)
-    }
+    ## otherwise...
+    ## hard-coded exceptions
     callName <- deparseOneLine(mainCall(playState)[[1]])
     if (callName %in% c("qqnorm", "qqplot")) {
         ## these return plotted coordinates in a list
@@ -593,19 +650,39 @@ xyData <- function(playState, space="plot")
         foo <- eval(modCall, playState$env)
         return(foo)
     }
-    x <- callArg(playState, 1)
-    y <- callArg(playState, "y")
-    xy.coords_with_class(playState, x, y)
+    ## check for named "data" argument
+    tmp.data <- callArg(playState, "data") ## may be NULL
+    ## look at first argument (tmp.data may be NULL)
+    tmp.x <- callArg(playState, 1, data=tmp.data)
+    if (inherits(tmp.x, "formula")) {
+        ## if 1st arg is formula, 2nd is `data`
+        if (is.null(tmp.data) &&
+            (is.null(names(mainCall)) ||
+             identical(names(mainCall)[[3]], "")))
+            tmp.data <- callArg(playState, 2)
+    }
+    if (is.null(tmp.data)) {
+        ## objects may also come from a with() block
+        if (identical(playState$call[[1]], as.symbol("with")))
+            tmp.data <- eval(playState$call[[2]], playState$env)
+    }
+    tmp.y <- callArg(playState, "y", data=tmp.data)
+    xy.coords_with_class(playState, tmp.x, tmp.y, data=tmp.data)
 }
 
 ## adapted from grDevices::xy.coords
-xy.coords_with_class <- function(playState, x, y=NULL, recycle=TRUE)
+xy.coords_with_class <- function(playState, x, y=NULL, recycle=TRUE, data=NULL)
 {
     if (is.null(y)) {
         if (is.language(x)) {
             if (inherits(x, "formula") && length(x) == 3) {
-                y <- eval(x[[2]], environment(x), playState$env)
-                x <- eval(x[[3]], environment(x), playState$env)
+                if (!is.null(data)) {
+                    y <- eval(x[[2]], data, playState$env)
+                    x <- eval(x[[3]], data, playState$env)
+                } else {
+                    y <- eval(x[[2]], environment(x), playState$env)
+                    x <- eval(x[[3]], environment(x), playState$env)
+                }
             }
             else stop("invalid first argument")
         }
